unit LogThreads;

interface

uses
  Windows, SysUtils, Classes;

const
  fmOpenCreate = $1000;
  fmOpenAlways = $2000;

type
  TLogStream = class(TStream)
  protected
    FHandle: THandle;
    procedure SetSize(NewSize: Longint); override;
  public
    constructor Create(const FileName: string; Mode: Word);
    destructor Destroy; override;

    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    procedure WriteStr(S: String); virtual;

    property Handle: THandle read FHandle;
  end;

  TPStrLink = ^TStrLink;
  TStrLink = record
    pNext: TPStrLink;
    S: String;
  end;

  TLogThread = class(TThread)
  protected
    FLogFile: TLogStream;

    pFirstStr, pLastStr: TPStrLink;
    hStrSem, hStrMutex: THandle;

    procedure Execute; override;
  public
    constructor Create(FileName: String); virtual;
    destructor Destroy; override;

    procedure WriteStr(S: String); virtual;
    property LogFile: TLogStream read FLogFile;
  end;





implementation

uses
  Consts;

{ TLogStream }

constructor TLogStream.Create(const FileName: string; Mode: Word);
var
  AccessMode, ShareMode, DispositionMode: LongWord;

begin
  FHandle := INVALID_HANDLE_VALUE;
  if Mode = fmCreate then Mode := fmOpenCreate or fmOpenReadWrite or fmShareDenyWrite;

  case Mode and $3 of
    fmOpenRead: AccessMode := GENERIC_READ;
    fmOpenWrite: AccessMode := GENERIC_WRITE;
    else AccessMode := GENERIC_READ or GENERIC_WRITE;
  end;
  case Mode and $70 of
    fmShareCompat: ShareMode := 0;
    fmShareExclusive: ShareMode := 0;
    fmShareDenyWrite: ShareMode := FILE_SHARE_READ;
    fmShareDenyRead: ShareMode := FILE_SHARE_WRITE;
    else ShareMode := FILE_SHARE_READ or FILE_SHARE_WRITE;
  end;
  case Mode and $3000 of
    fmOpenCreate: DispositionMode := CREATE_NEW;
    fmOpenCreate or fmOpenAlways: DispositionMode := CREATE_ALWAYS;
    0: DispositionMode := OPEN_EXISTING;
    else DispositionMode := OPEN_ALWAYS;
  end;

  FHandle := CreateFile(PChar(FileName), AccessMode, ShareMode, nil, DispositionMode, FILE_ATTRIBUTE_NORMAL, 0);

  if FHandle = INVALID_HANDLE_VALUE then begin
    if (Mode and fmOpenCreate) <> 0 then
      raise EFCreateError.CreateFmt(SFCreateError, [FileName])
    else
      raise EFOpenError.CreateFmt(SFOpenError, [FileName])
    ;
  end;
end;

destructor TLogStream.Destroy;
begin
  if FHandle <> INVALID_HANDLE_VALUE then FileClose(FHandle);
end;

function TLogStream.Read(var Buffer; Count: Integer): Longint;
begin
  if not ReadFile(Handle, Buffer, Count, LongWord(Result), nil) then
    Result := 0
  ;
end;

function TLogStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  Result := FileSeek(FHandle, Offset, Origin);
end;

procedure TLogStream.SetSize(NewSize: Integer);
begin
  Seek(NewSize, soFromBeginning);
  Win32Check(SetEndOfFile(FHandle));
end;

function TLogStream.Write(const Buffer; Count: Integer): Longint;
begin
  if not WriteFile(Handle, Buffer, Count, LongWord(Result), nil) then
    Result := 0
  ;
end;

procedure TLogStream.WriteStr(S: String);
begin
  WriteBuffer(S[1], Length(S));
end;

{ TLogThread }

constructor TLogThread.Create(FileName: String);
begin
  inherited Create(True);

  FLogFile := TLogStream.Create(FileName, fmOpenCreate or fmOpenAlways or fmOpenReadWrite or fmShareDenyNone);
  hStrSem := CreateSemaphore(nil, 0, $FFFFFF, nil);
  hStrMutex := CreateMutex(nil, False, nil);

  Resume;
end;

destructor TLogThread.Destroy;
begin
  //Send an extra "kill release" to the semaphore, without a corresponding link in the list.
  ReleaseSemaphore(hStrSem, 1, nil);

  inherited;

  FLogFile.Free;
  CloseHandle(hStrSem);
  CloseHandle(hStrMutex);
end;

procedure TLogThread.Execute;
var
  pStrLink: TPStrLink;
  //BufStr: String;

begin
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_BELOW_NORMAL);
  try
    //BufStr := '';

    repeat
      WaitForSingleObject(hStrSem, INFINITE);

      WaitForSingleObject(hStrMutex, INFINITE);
        pStrLink := pFirstStr;
        if pStrLink = nil then begin
        end
        else if pStrLink.pNext = nil then begin
          pFirstStr := nil;
          pLastStr := nil;
        end
        else begin
          pFirstStr := pStrLink.pNext;
        end;
      ReleaseMutex(hStrMutex);

      if pStrLink = nil then Break; //It was a "kill release".

      LogFile.WriteStr(pStrLink.S);
      //BufStr := BufStr + pStrLink.S;
      Dispose(pStrLink);

      {
      if Length(BufStr) > 7*1024 then begin
        LogFile.WriteStr(BufStr);
        BufStr := '';
      end;
      }
    until False;

    //if BufStr <> '' then LogFile.WriteStr(BufStr);
  except
    on E: Exception do MessageBox(0, PChar(E.Message), PChar(String(E.ClassName)), MB_ICONEXCLAMATION);
  end;
end;

procedure TLogThread.WriteStr(S: String);
var
  pStrLink: TPStrLink;

begin
  New(pStrLink);
  pStrLink.pNext := nil;
  pStrLink.S := S;

  WaitForSingleObject(hStrMutex, INFINITE);
    //Add it to the linked list.
    if pLastStr <> nil then
      pLastStr.pNext := pStrLink
    else
      pFirstStr := pStrLink
    ;
    pLastStr := pStrLink;
  ReleaseMutex(hStrMutex);

  ReleaseSemaphore(hStrSem, 1, nil);
end;

end.
